from typing import Union
from .currency import Currency

Addable = Union['CurrencyRange', Currency]


class CurrencyRange:
    """Taxable currency range"""

    __slots__ = ('start', 'stop'
                 )

    def __init__(self, start: Currency, stop: Currency) -> None:
        if not isinstance(start, Currency) and not isinstance(stop, Currency):
            raise TypeError(
                f'Currency Range start and stop need to be of type Currency'
            )
        if start.currency != stop.currency:
            raise ValueError(
                f'Cannot create range for different currencies. start:{start.currency} stop:{stop.currency}'
            )
        if start > stop:
            raise ValueError(
                f'Cannot create a range from {start.amount} to {stop.amount}'
            )
        self.start = start
        self.stop = stop

    def __str__(self) -> str:
        return f'CurrencyRange({self.start} {self.stop})'

    @property
    def currency(self) -> str:
        """ return the currency denomination. Like USD or EUR """
        return self.start.currency

    def __add__(self, other: Addable) -> 'CurrencyRange':
        if isinstance(other, Currency):
            if self.currency != other.currency:
                raise ValueError(
                    f'cannot add ranges in {self.currency} and {other.currency}'
                )
            start = self.start + other
            stop = self.stop + other
            return CurrencyRange(start, stop)
        elif isinstance(other, CurrencyRange):
            if other.start.currency != self.currency:
                raise ValueError(
                    f'cannot add ranges in {self.currency} and {other.start.currency}'
                )
            start = self.start + other.start
            stop = self.stop + other.stop
            return CurrencyRange(start, stop)
        return NotImplemented

    def __sub__(self, other: Addable) -> 'CurrencyRange':
        if isinstance(other, Currency):
            if self.currency != other.currency:
                raise ValueError(
                    f'cannot add ranges in {self.currency} and {other.currency}'
                )
            start = self.start - other
            stop = self.stop - other
            return CurrencyRange(start, stop)
        elif isinstance(other, CurrencyRange):
            if other.start.currency != self.currency:
                raise ValueError(
                    f'cannot add ranges in {self.currency} and {other.start.currency}'
                )
            start = self.start - other.start
            stop = self.stop - other.stop
            return CurrencyRange(start, stop)
        return NotImplemented

    def __eq__(self, other: object) -> bool:
        if isinstance(other, CurrencyRange):
            return (
                self.start == other.start and
                self.stop == other.stop)
        return False

    def __contains__(self, item: Currency) -> bool:
        if not isinstance(item, Currency):
            raise TypeError(
                f'currency_range requires currency as left operand, not {type(item)}')
        return self.start <= item <= self.stop

    def quantize(self, exp=None, rounding=None) -> 'CurrencyRange':
        """Return a copy of the range with start and stop quantized.

        All arguments are passed to 'Currency.quantize'.
        """
        return CurrencyRange(
            self.start.quantize(exp, rounding=rounding),
            self.stop.quantize(exp, rounding=rounding))

    def replace(self, start: Currency = None, stop: Currency = None) -> 'CurrencyRange':
        """Return a range with start or stop replaced with given values."""
        if start is None:
            start = self.start
        if stop is None:
            stop = self.stop
        return CurrencyRange(start=start, stop=stop)



import pytest

from tekmoney.currency import Currency
from tekmoney.currency_tax import CurrencyWithTax
from tekmoney.utils import tek_sum


def test_init():
    currency = CurrencyWithTax(net=Currency(1, 'USD'), gross=Currency(1, 'USD'))
    assert (currency.net == Currency(1, 'USD')) and (currency.gross == Currency(1, 'USD'))
    with pytest.raises(ValueError):
        CurrencyWithTax(net=Currency(1, 'USD'), gross=Currency(1, 'GBP'))
    with pytest.raises(TypeError):
        CurrencyWithTax(1, 1)


def test_str():
    currency = CurrencyWithTax(net=Currency(1, 'USD'), gross=Currency(1, 'USD'))
    assert str(currency) == 'CurrencyWithTax(net=1 USD, gross=1 USD)'


def test_add():
    currency1 = CurrencyWithTax(Currency(10, 'USD'), Currency(15, 'USD'))
    currency2 = CurrencyWithTax(Currency(20, 'USD'), Currency(30, 'USD'))
    assert currency2 + currency1 == CurrencyWithTax(Currency(30, 'USD'), Currency(45, 'USD'))
    currency3 = currency1 + Currency(5, 'USD')
    assert currency3.net == Currency(15, 'USD')
    assert currency3.gross == Currency(20, 'USD')
    with pytest.raises(ValueError):
        currency1 + CurrencyWithTax(Currency(10, 'GBP'), Currency(10, 'GBP'))
    with pytest.raises(TypeError):
        currency1 + 1


def test_sub():
    currency1 = CurrencyWithTax(Currency(10, 'USD'), Currency(15, 'USD'))
    currency2 = CurrencyWithTax(Currency(30, 'USD'), Currency(45, 'USD'))
    assert currency2 - currency1 == CurrencyWithTax(Currency(20, 'USD'), Currency(30, 'USD'))
    currency3 = currency1 - Currency(5, 'USD')
    assert currency3.net == Currency(5, 'USD')
    assert currency3.gross == Currency(10, 'USD')
    with pytest.raises(ValueError):
        currency1 - CurrencyWithTax(Currency(10, 'GBP'), Currency(10, 'GBP'))
    with pytest.raises(TypeError):
        currency1 - 1


def test_mult():
    currency1 = CurrencyWithTax(Currency(10, 'EUR'), Currency(20, 'EUR'))
    assert currency1 * 2 == CurrencyWithTax(Currency(20, 'EUR'), Currency(40, 'EUR'))
    assert 2 * currency1 == currency1 * 2
    with pytest.raises(TypeError):
        currency1 * currency1


def test_div():
    currency1 = CurrencyWithTax(Currency(10, 'EUR'), Currency(20, 'EUR'))
    assert currency1 / 2 == CurrencyWithTax(Currency(5, 'EUR'), Currency(10, 'EUR'))
    with pytest.raises(TypeError):
        currency1 / currency1


def test_comparison():
    currency = CurrencyWithTax(Currency(10, 'EUR'), Currency(15, 'EUR'))
    assert currency == CurrencyWithTax(Currency(10, 'EUR'), Currency(15, 'EUR'))
    assert currency != CurrencyWithTax(Currency(20, 'EUR'), Currency(30, 'EUR'))
    assert currency != CurrencyWithTax(Currency(10, 'GBP'), Currency(15, 'GBP'))
    assert currency != Currency(10, 'EUR')
    assert currency < CurrencyWithTax(Currency(20, 'EUR'), Currency(30, 'EUR'))
    assert currency <= CurrencyWithTax(Currency(10, 'EUR'), Currency(15, 'EUR'))
    assert currency <= CurrencyWithTax(Currency(20, 'EUR'), Currency(30, 'EUR'))
    assert currency > CurrencyWithTax(Currency(1, 'EUR'), Currency(1, 'EUR'))
    assert currency >= CurrencyWithTax(Currency(10, 'EUR'), Currency(15, 'EUR'))
    assert currency >= CurrencyWithTax(Currency(1, 'EUR'), Currency(1, 'EUR'))
    assert not currency <= CurrencyWithTax(Currency(1, 'EUR'), Currency(1, 'EUR'))
    assert not currency >= CurrencyWithTax(Currency(20, 'EUR'), Currency(30, 'EUR'))
    with pytest.raises(ValueError):
        currency < CurrencyWithTax(Currency(10, 'GBP'), Currency(15, 'GBP'))
    with pytest.raises(TypeError):
        currency >= Currency(1, 'EUR')
    with pytest.raises(TypeError):
        currency < Currency(20, 'EUR')


def test_quantize():
    price = CurrencyWithTax(Currency('1.001', 'EUR'), Currency('1.001', 'EUR'))
    assert price.quantize() == (
        CurrencyWithTax(Currency('1.00', 'EUR'), Currency('1.00', 'EUR')))


def test_currency():
    price = CurrencyWithTax(Currency(1, 'PLN'), Currency(1, 'PLN'))
    assert price.currency == 'PLN'


def test_tax():
    price = CurrencyWithTax(Currency(10, 'USD'), Currency(15, 'USD'))
    assert price.tax == Currency(5, 'USD')


def test_tek_sum():
    assert tek_sum([Currency(5, 'USD'), Currency(10, 'USD')]) == Currency(15, 'USD')
    with pytest.raises(TypeError):
        tek_sum([])



import pytest

from tekmoney.currency import Currency
from tekmoney.currency_range import CurrencyRange


def test_construction():
    price1 = Currency(10, 'EUR')
    price2 = Currency(30, 'EUR')
    price_range = CurrencyRange(price1, price2)
    assert price_range.start == price1
    assert price_range.stop == price2
    with pytest.raises(ValueError):
        CurrencyRange(price1, Currency(20, 'PLN'))
    with pytest.raises(ValueError):
        CurrencyRange(price2, price1)


def test_addition_with_money():
    price1 = Currency(10, 'EUR')
    price2 = Currency(30, 'EUR')
    price_range = CurrencyRange(price1, price2)
    price3 = Currency(40, 'EUR')
    result = price_range + price3
    assert result.start == price1 + price3
    assert result.stop == price2 + price3
    with pytest.raises(ValueError):
        price_range + Currency(1, 'BTC')


def test_addition_with_money_range():
    price1 = Currency(10, 'EUR')
    price2 = Currency(30, 'EUR')
    price_range1 = CurrencyRange(price1, price2)
    price3 = Currency(40, 'EUR')
    price4 = Currency(80, 'EUR')
    price_range2 = CurrencyRange(price3, price4)
    result = price_range1 + price_range2
    assert result.start == price1 + price3
    assert result.stop == price2 + price4
    with pytest.raises(ValueError):
        price_range1 + CurrencyRange(Currency(1, 'BTC'), Currency(2, 'BTC'))


def test_addition_with_other_types():
    price1 = Currency(10, 'EUR')
    price2 = Currency(30, 'EUR')
    price_range = CurrencyRange(price1, price2)
    with pytest.raises(TypeError):
        price_range + 1


def test_subtraction_with_money():
    price1 = Currency(40, 'EUR')
    price2 = Currency(80, 'EUR')
    price_range = CurrencyRange(price1, price2)
    price3 = Currency(10, 'EUR')
    result = price_range - price3
    assert result.start == price1 - price3
    assert result.stop == price2 - price3
    with pytest.raises(ValueError):
        price_range - Currency(1, 'BTC')


def test_subtraction_with_money_range():
    price1 = Currency(10, 'EUR')
    price2 = Currency(30, 'EUR')
    price_range1 = CurrencyRange(price1, price2)
    price3 = Currency(40, 'EUR')
    price4 = Currency(80, 'EUR')
    price_range2 = CurrencyRange(price3, price4)
    result = price_range2 - price_range1
    assert result.start == price3 - price1
    assert result.stop == price4 - price2
    with pytest.raises(ValueError):
        price_range2 - CurrencyRange(Currency(1, 'BTC'), Currency(2, 'BTC'))


def test_subtraction_with_other_types():
    price1 = Currency(40, 'EUR')
    price2 = Currency(80, 'EUR')
    price_range = CurrencyRange(price1, price2)
    with pytest.raises(TypeError):
        price_range - 1


def test_comparison():
    price1 = Currency(10, 'EUR')
    price2 = Currency(30, 'EUR')
    price_range1 = CurrencyRange(price1, price2)
    price3 = Currency(40, 'EUR')
    price4 = Currency(80, 'EUR')
    price_range2 = CurrencyRange(price3, price4)
    assert price_range1 == CurrencyRange(price1, price2)
    assert price_range1 != price_range2
    assert price_range1 != CurrencyRange(price1, price1)
    assert price_range1 != CurrencyRange(
        Currency(10, 'USD'), Currency(30, 'USD'))
    assert price_range1 != price1


def test_membership():
    price1 = Currency(10, 'EUR')
    price2 = Currency(30, 'EUR')
    price_range = CurrencyRange(price1, price2)
    assert price1 in price_range
    assert price2 in price_range
    assert (price1 + price2) / 2 in price_range
    assert price1 + price2 not in price_range
    with pytest.raises(TypeError):
        15 in price_range


def test_quantize():
    price1 = Currency(10, 'EUR')
    price2 = Currency(30, 'EUR')
    price_range = CurrencyRange(price1, price2)
    result = price_range.quantize()
    assert str(result.start.amount) == '10.00'
    assert str(result.stop.amount) == '30.00'


def test_replace():
    price1 = Currency(10, 'EUR')
    price2 = Currency(30, 'EUR')
    price3 = Currency(20, 'EUR')
    price_range = CurrencyRange(price1, price2)
    result = price_range.replace(stop=price3)
    assert result.start == price1
    assert result.stop == price3
    result = price_range.replace(start=price3)
    assert result.start == price3
    assert result.stop == price2


def test_currency():
    price1 = Currency(10, 'EUR')
    price2 = Currency(30, 'EUR')
    price_range = CurrencyRange(price1, price2)
    assert price_range.currency == 'EUR'


def test_str():
    price1 = Currency(10, 'EUR')
    price2 = Currency(30, 'EUR')
    price_range = CurrencyRange(price1, price2)
    assert str(price_range) == (
        "CurrencyRange(10 EUR 30 EUR)")
